#pragma once

namespace dolfinx
{
/*! \namespace dolfinx::generation
    \brief Generators for simple meshes

    Generators for structured meshes for simple shapes.
*/
}

// DOLFINX mesh generation interface

#include <dolfinx/generation/BoxMesh.h>
#include <dolfinx/generation/IntervalMesh.h>
#include <dolfinx/generation/RectangleMesh.h>
