#pragma once

namespace dolfinx
{
/*! \namespace dolfinx::graph
    \brief Graph data structures and algorithms

    Data structures for building and representing graphs, and algorithms on
   graphs, e.g., re-ordering and partitioning.
*/
}

// DOLFINX graph interface

#include <dolfinx/graph/boostordering.h>
#include <dolfinx/graph/partition.h>
