#pragma once

// DOLFINX interface

#include <dolfinx/common/dolfin_common.h>
#include <dolfinx/fem/dolfin_fem.h>
#include <dolfinx/generation/dolfin_generation.h>
#include <dolfinx/geometry/dolfin_geometry.h>
#include <dolfinx/graph/dolfin_graph.h>
#include <dolfinx/io/dolfin_io.h>
#include <dolfinx/la/dolfin_la.h>
#include <dolfinx/mesh/dolfin_mesh.h>
#include <dolfinx/nls/dolfin_nls.h>
#include <dolfinx/refinement/dolfin_refinement.h>
