#pragma once

namespace dolfinx
{
/*! \namespace dolfinx::refinement
    \brief Mesh refinement algorithms

    Methods for refining meshes uniformly, or with markers, using
    edge bisection.
*/
}

// DOLFINX refinement interface

#include <dolfinx/refinement/refine.h>
