// Copyright (C) 2003-2006 Anders Logg
//
// This file is part of DOLFINX (https://www.fenicsproject.org)
//
// SPDX-License-Identifier:    LGPL-3.0-or-later

/*! \mainpage DOLFIN-X Programmer's Reference
 *
 * This is the documentation of the DOLFIN-X C++ interface.
 *
 */
