// Copyright (C) 2019 Chris Richardson
//
// This file is part of DOLFINX (https://www.fenicsproject.org)
//
// SPDX-License-Identifier:    LGPL-3.0-or-later

#pragma once

#define LOGURU_WITH_STREAMS 1
#define LOGURU_REPLACE_GLOG 1

#include "loguru.hpp"
