#pragma once

namespace dolfinx
{
/*! \namespace dolfinx::common
    \brief Miscellaneous classes, functions and types

    This namespace provides utility type functions for managing subsystems,
   convenience classes and library-wide typedefs.
*/
}

// DOLFINX common

#include <dolfinx/common/MPI.h>
#include <dolfinx/common/subsystem.h>
#include <dolfinx/common/Table.h>
#include <dolfinx/common/Timer.h>
#include <dolfinx/common/defines.h>
#include <dolfinx/common/init.h>
#include <dolfinx/common/timing.h>
#include <dolfinx/common/types.h>
#include <dolfinx/common/version.h>
