#pragma once

namespace dolfinx
{
/*! \namespace dolfinx::mesh
    \brief Mesh data structures and algorithms on meshes

    Representations of meshes and support for operations on meshes.
*/
}

// DOLFINX mesh interface

#include <dolfinx/graph/AdjacencyList.h>
#include <dolfinx/mesh/Geometry.h>
#include <dolfinx/mesh/Mesh.h>
#include <dolfinx/mesh/MeshTags.h>
#include <dolfinx/mesh/Topology.h>
#include <dolfinx/mesh/cell_types.h>
#include <dolfinx/mesh/utils.h>
