#pragma once

namespace dolfinx
{
/*! \namespace dolfinx::nls
    \brief Nonlinear solvers

    Methods for solving nonlinear equations.
*/
}

// DOLFINX nonlinear solver

#include <dolfinx/nls/NewtonSolver.h>
