#pragma once

namespace dolfinx
{
/*! \namespace dolfinx::geometry
    \brief Geometry data structures and algorithms

    Tools for geometric data structures and operations, e.g. searching.
*/
}

// DOLFINX geometry interface

#include <dolfinx/geometry/BoundingBoxTree.h>
#include <dolfinx/geometry/gjk.h>
